/* ========================================================================
 * Copyright (c) 2011 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

/* gwsd ding responder */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <pthread.h>

#include "gwsd.h"

#define DING_RECV_LEN 24
#define DING_PORT 10101

extern int verbose;

int recvsock;
int sendsock;
int proxd_sock;
struct sockaddr_in proxd_addr;

char my_hostname[64];
char *my_proxd_app = "gwsd";

int nproxderr = 0;
int reader_running = 0;

/* This is external, filled by various processes */
char ok_status[256];
char err_status[1024];
int err_level = 0;

/* Thread to receive dings */

void *th_ding_recv(void *arg)
{
   int fd = recvsock;
   char dgram[DING_RECV_LEN+1];
   struct sockaddr_in from;
   socklen_t fromlen = sizeof(from);
   int r;
   unsigned char *b;
   char *msg;

   syslog (LOG_DEBUG, "ding reader starting on fd %d", fd);
   reader_running = 1;
   while ((r=recvfrom(fd, dgram, DING_RECV_LEN, 0,
             (struct sockaddr *)&from,&fromlen))>0) {
      if (r<0) {
         syslog(LOG_ERR, "ding recv %m");
         sleep (5);
         continue;
      }
      dgram[r] = '\0';
      // syslog(LOG_DEBUG, "recv:%s\n", dgram);
      if (isupper(*dgram)) *dgram = tolower(*dgram);
  
      msg = NULL;
      if (*dgram=='s') {
         if (err_level) msg = err_status;
         else {
            sprintf(ok_status, "0p %d clients", num_client);
            msg = ok_status;
         }
      } else if (*dgram=='r') {
         err_level = 0;  // reset error
         msg = "0preset";
      }
  
      if (msg) {
         int msgl = strlen(msg);
         if ((r = sendto(fd, msg, msgl, 0, (const struct sockaddr*)&from, sizeof(from))) != msgl) {
             syslog(LOG_ERR, "ding send %m");
             sleep(5);
         }
         // b = (unsigned char *) &from.sin_addr.s_addr;
         // syslog(LOG_DEBUG, "Repl to %d.%d.%d.%d (%d).", b[0],b[1],b[2],b[3],ntohs(from.sin_port));
      }

   }
   syslog(LOG_ALERT,"ding reader exiting");
   reader_running = 0;
}


int ding_init(GWSConfig *cfg)
{
   pthread_t thread;
   pthread_attr_t pta;
   struct sockaddr_in dgaddr;
   struct hostent *host;
   struct servent *sp;

   int fc;
   int port;

   gethostname(my_hostname, 64);
   strcpy(ok_status, "0pInitialized");

   /* Initialize the ding and proxd systems */

   if (cfg->ding_service) {

      sp = getservbyname(cfg->ding_service, "udp");
      if (!sp) {
          syslog(LOG_ERR, "couldn't get service for %s: %m", cfg->ding_service);
          return (0);
      }
      syslog(LOG_INFO, "Ding listener on port %d", ntohs(sp->s_port));

      /* datagram receiver */
      recvsock = socket(AF_INET, SOCK_DGRAM, 17);
      if (recvsock < 0) {
        syslog(LOG_ERR, "ding socket %m");
        return (0);
      }
      dgaddr.sin_family = AF_INET;
      dgaddr.sin_addr.s_addr = INADDR_ANY;
      dgaddr.sin_port = sp->s_port;
      if (bind(recvsock, (const struct sockaddr*)&dgaddr, sizeof(dgaddr))) {
         syslog(LOG_ERR, "binding ding socket %m");
         return (0);
      }

      pthread_attr_init(&pta);
      pthread_attr_setdetachstate(&pta, PTHREAD_CREATE_DETACHED);
      pthread_create(&thread, &pta, th_ding_recv, (void*) NULL);


   } else {
      syslog(LOG_WARNING, "No ding system configured.  Dings will not be processed.");
   }


   /* initialize the proxd sender */

   if (cfg->proxd_host) {

      proxd_sock = 0;
      proxd_addr.sin_family = AF_INET;
      host = gethostbyname(cfg->proxd_host);
      if (!host) {
          syslog(LOG_ERR, "couldn't get host for %s: %m", cfg->proxd_host);
          return (0);
      }
      proxd_addr.sin_family = host->h_addrtype;
      bcopy(host->h_addr, &proxd_addr.sin_addr, host->h_length);
 
      sp = getservbyname(cfg->proxd_service, "udp");
      if (!sp) {
          syslog(LOG_ERR, "couldn't get service for %s: %m", cfg->proxd_service);
          return (0);
      }
      proxd_addr.sin_port = sp->s_port;
    
      if ((proxd_sock = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
         syslog(LOG_ERR, "couldn't get proxd socket: %m");
         perror("socket");
         return (0);
      }
  } else {
     syslog(LOG_WARNING, "No proxd logger configured.  Messages will not be sent to proxd.");
  }

   return (1);
}
   

/* proxd reporting */

static char *proxd_xml_template = "<?xml version=\"1.0\" ?><transactionlist><transaction>%s</transaction></transactionlist>";

static char *proxd_upd_template = "<?xml version=\"1.0\" ?><transactionlist><transaction>\
<newAlert host=\"%s\" component=\"%s\" sev=\"%d\" msg=\"%s\" longmsg=\"%s\" contact=\"oncall\" count=\"+\"/>\
</transaction></transactionlist>";


/* report errors to DS-UNIX error system */

int report_error(int level, char *ext, char *smsg, char *lmsg) 
{
   char buf[1024];
   char zapp[256];
   int nb;
  
   if (verbose>1) syslog(LOG_ERR, "reporting error: %s", smsg);

   if (level>err_level) {
      if (!lmsg) lmsg = "";
      snprintf(err_status, 1024, "%dp %s: %s. %s", level, ext, smsg, lmsg);
      err_level = level;
   }

   if (!proxd_sock) return (0);

   if (ext) snprintf(zapp, 256, "%s (%s)", my_proxd_app, ext);
   else strcpy(zapp, my_proxd_app);

   snprintf(buf, 1024, proxd_upd_template, my_hostname, zapp, level, smsg, lmsg?lmsg:"");
   nb = strlen(buf);
   if (sendto(proxd_sock,buf,nb,0,(struct sockaddr *)&proxd_addr,sizeof(proxd_addr)) != nb) {
      syslog(LOG_ERR, "sending proxd message: %m");
      if (nproxderr++ > 10) {
         syslog(LOG_ERR, "shutting down proxd messaging");
         proxd_sock = 0;
      }
      return (0);
   }
   return (1);
}

