/* ========================================================================
 * Copyright (c) 2011 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

/* crypt routines for mod_gws */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#define UC (unsigned char*)

#include "openssl/buffer.h"
#include "openssl/bio.h"
#include "openssl/evp.h"
#include "gwscrypt.h"

static EVP_CIPHER_CTX *ct[2] = {NULL,NULL};

/* augment srypt seed */
void gws_crypt_seed()
{
   struct timeval tv;
   gettimeofday(&tv,NULL);
   RAND_seed(&tv.tv_usec, 4);
}

/* Initialize. Set password for cryptions */
int gws_crypt_init(char *keytext, int keylen)
{
   const EVP_CIPHER *c;
   const EVP_MD *m;
   int i;
   unsigned char key[EVP_MAX_KEY_LENGTH],iv[EVP_MAX_KEY_LENGTH];
   char text[32];

   if (ct[0]) return(1);

   /* printf("GWS: crypt init\n"); */

   gws_crypt_seed();
   c = EVP_aes_128_ecb();  /* AES */
   m = EVP_dss();
   if (!(c&&m)) return (0);

   gws_crypt_seed();
   RAND_bytes(text, 32);
   EVP_BytesToKey(c, m, NULL, UC keytext, keylen, 1, key, iv);

   for (i=0;i<2;i++) {
     ct[i]=(EVP_CIPHER_CTX *)malloc(sizeof(EVP_CIPHER_CTX));
     EVP_CipherInit(ct[i], c, key, iv, i);
   }
   return (1);
}

   
/* Base64 encode */
int gws_data_to_b64(char **d64, int *d64l, char *data, int dl)
{
   BIO *b64, *bmem;
   BUF_MEM *bmem_mem;
   int i;

   b64 = BIO_new(BIO_f_base64());
   bmem = BIO_new(BIO_s_mem());
   BIO_push(b64, bmem);

   BIO_write(b64, data, dl);
   BIO_flush(b64);
   BIO_get_mem_ptr(bmem, &bmem_mem); 
 // fprintf(stderr, "outlen: %d\n", bmem_mem->length);
 // fprintf(stderr, "outtxt: [%s]\n", bmem_mem->data);

   // strip the newlines
   char *z = (char*)malloc( bmem_mem->length+1);
   *d64 = z;
   for (i=0;i<bmem_mem->length;i++) if (bmem_mem->data[i]!='\n') *z++ = bmem_mem->data[i];
   *z = '\0';
 // fprintf(stderr, "outnnl: [%s]\n", *d64);
   if (d64l) *d64l = i;
   BIO_free_all(bmem);
   return (1);
}

/* Base64 decode */
int gws_b64_to_data(char **data, int *dl, char *d64, int d64l)
{
   char *buf;
   int nb;
   BIO *b64, *bmem;
   char *d64nl = (char*) malloc(d64l + d64l/64 + 2);
   char *p;
   int i = 0;

   // insert newlines
   p = d64nl;
   while (i<d64l) {
      *p++ = d64[i++];
      if ((i/64)*64 == i) *p++ = '\n';
   }
   if (*(p-1)!='\n') *p++ = '\n';
   *p = '\0';
   d64l = strlen(d64nl);

 // fprintf(stderr, "innnl: [%s]\n", d64nl);

   b64 = BIO_new(BIO_f_base64());
   bmem = BIO_new_mem_buf(d64nl, d64l);
   BIO_push(b64, bmem);

   buf = malloc(d64l*2);
   nb = BIO_read(b64, buf, d64l);
   BIO_flush(b64);
   *data = buf;

 // fprintf(stderr, "outtxt: [%s]\n", buf);

   if (dl) *dl = nb;
   BIO_free_all(bmem);
   free(d64nl);
   return (1);
}


/* Crypt routine:
   mode = GWS_ENCRYPT:  plaintext (in) -> ciphertext (out)
   mode = GWS_DECRYPT:  ciphertext (in) -> plaintext (out)
   */

int gws_crypt(int mode, char **out, int *outlen, char *in, int inlen)
{
   int ol;
   int len;
   char *buf = (char*) malloc(inlen*3);
   char *bufp = buf;
   int s;

   /* PRINTF("gws crypt pid (%d) %d:%x\n", getpid(), mode, ct[0]); */
   EVP_CipherInit(ct[mode], NULL, NULL, NULL, mode);
   s = EVP_CipherUpdate(ct[mode], UC bufp, &ol, UC in, inlen);
   bufp += ol;
   len = ol;

   if (s) {
      EVP_CipherFinal(ct[mode], UC bufp, &ol);
      len += ol;
   }

   *out = buf;
   if (outlen) *outlen = len;
   return (len);
}

// generate a fnv hash of a string
#define FNV1_32_INIT ((u_int)0x811c9dc5)
#define FNV_32_PRIME ((u_int)0x01000193)

u_int gws_crypt_hash(char *str) 
{
   u_int hval = FNV1_32_INIT;
   unsigned char *s = (unsigned char *)str;
   while (*s) {
     hval *= FNV_32_PRIME;
     hval ^= *s++;
   }
   return hval;
}
