/* ========================================================================
 * Copyright (c) 2011 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

/*
   GWS daemon for mod_gws 
  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <ctype.h>
#include <strings.h>
#include <errno.h>
#include <pthread.h>
 
#include "gwsd.h"
#include "gwslib.h"

#define LOG syslog
int log_facility = LOG_SYSLOG;

extern int      errno;
int hupped = 0;
 
#define VERSION     "1.0"

int    verbose = 0; 

char    *my_name;
int seq = 0;
time_t gwsd_started;
 
char *config_name;

GWSConfig *gws_config;
int funix = (-1);  // our listener socket
int pidfd = (-1);  // pid file

/* Generic (maybe do something) exit. */

void err_exit() {
    LOG(LOG_INFO, "Exiting");
    if (funix>=0) {
       close(funix);
       unlink(gws_config->socket_path);
    }
    if (pidfd>=0) {
       close(pidfd);
       unlink(gws_config->pid_file);
    }
    exit (1);
}
 
/* Signal handler */

static sigset_t sigsmask; 

int sigs_to_catch[] = {SIGHUP, SIGUSR1, SIGPIPE, SIGINT, SIGTERM, SIGQUIT, 0};

void handle_signals(int sig) {
   switch (sig) {
     case SIGUSR1: 
     case SIGHUP:  hupped = 1;     /* set flag */
                   break;

     case SIGPIPE: break;

     default:      err_exit();
   }
} 
  

static struct timeval utime;
long time_mark()
{
   gettimeofday(&utime,0);
   return (utime.tv_sec);
}


/** Locking. */

typedef pthread_mutex_t LOCK_T;
#define LOCK(l)   pthread_mutex_lock(&l)
#define UNLOCK(l) pthread_mutex_unlock(&l)
#define NOCANCEL pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
#define OKCANCEL pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
#define CANCEL(th) pthread_cancel(th)

void *th_client_handler(void *arg);


char *MALLOC(int siz, char *memo)
{
  char *m;
  m = (char*) malloc(siz);
  if (m) return(m);
  perror ("MALLOC");
  *m = 0;
  return (NULL);
}

char *STRDUP(char *str, char *memo)
{
   char *s = MALLOC(strlen(str)+1, memo);
   strcpy(s,str);
   return (s);
}

void FREE(char *p, char *memo)
{
  if (!p) return;
  free(p);
}


/** Client and server data.  */

#define SMSG_LIST_SIZE 1024

typedef struct Client__ {
  int seq;
  int  fd;                      /* File descriptor for this client  */
  int pwok;                     /* client has provided password */
  pthread_t handler; 
} Client_, *Client;

LOCK_T client_lock;
LOCK_T gws_lock;
int num_client = 0;

/** Allocate a client struct. */

Client new_client()
{
   Client c;
   c = (Client) MALLOC(sizeof(Client_), "newcl");
   memset(c, '\0', sizeof(Client_));
   num_client++;
   return (c);
}

void free_client(Client c)
{
   free(c);
   num_client--;
}


// returns 1 on complete message
int recv_message(int fd, void *buf, int nb)
{
   int n = nb;
   int r;
   void *b = buf;
   while ((r=read(fd, b, n)) > 0 ) {
      if (r == n) return (1);
      b += r;
      n -= r;
   }
   LOG(LOG_INFO,"get %d bytes: fd=%d, err=%d", nb, fd, errno);
   return (0);
}


/** --- Client handler --- **/

/* Check an is-member request */

int process_member_of(Client C, ClientMsg msg, ResponseMsg resp)
{
   if (!C->pwok) return (401);
   LOCK(gws_lock);
   resp->resp = gwslib_is_member(gws_config, msg->group, msg->id, msg->actas);
   UNLOCK(gws_lock);
   LOG(LOG_DEBUG, "test %s member of %s, status: %d", msg->id, msg->group, resp->resp);
   return (resp->resp);
}

/* Check a password */

int process_password(Client C, ClientMsg msg, ResponseMsg resp)
{
   if (!strcmp(msg->id, gws_config->client_pw)) {
      resp->resp = 200;
      C->pwok = 1;
   } else {
      report_error(2, "gws", "client password failed", NULL);
      resp->resp = 401;
   }
   return (resp->resp);
}


/* Client handler. Receive and process messages. */

void *th_client_handler(void *arg)
{
   Client C = (Client) arg;
   ClientMsg_ msg;
   ResponseMsg_ resp;
   int nb;

   int fd = C->fd;

   while (recv_message(fd, (void*)&msg, sizeof(ClientMsg_))) {
      
      resp.seq = msg.seq;
      resp.resp = RESP_INVALID;
      if (msg.cmd == CMD_MEMBER_OF) resp.resp = process_member_of(C, &msg, &resp);
      else if (msg.cmd == CMD_PASSWORD) resp.resp = process_password(C, &msg, &resp);
      nb = write(C->fd, &resp, sizeof(resp));
      if (nb != sizeof(resp)) {
         LOG(LOG_ERR, "Error %d responding to client %d", errno, C->seq);
         break;
      }
      LOG(LOG_DEBUG, "Responded %d to client %d(%d)", resp.resp, C->seq, resp.seq);
   }

   if (verbose) LOG(LOG_DEBUG, "[%d] handler thread exit (fd=%d)\n", C->seq, C->fd);
   if (fd) close(fd);
   free_client(C);
   return (NULL);
}

/* Config is name value pairs.  This is read before we are daemon, so can write to stderr also. */


GWSConfig *getconfig(char *filename)
{
   char buf[1024];

   GWSConfig *cfg = (GWSConfig*) malloc(sizeof(GWSConfig));
   memset(cfg, '\0', sizeof(GWSConfig));
   cfg->client_pw = (char*) malloc(64);
   *cfg->client_pw = '\0';
   cfg->actas_hdr = "X-UW-Act-as";  // uw default
   cfg->socket_path = "/var/run/gwsd.sock";  // ditto
   cfg->pid_file = "/var/run/gwsd.pid";      // ditto
   
   FILE *f = fopen(filename, "r");
   if (!f) {
      syslog(LOG_ERR, "opening %s: %m", filename);
      perror(filename);
      return (NULL);
   }
   while (fgets(buf, 512, f)) {
      char *k, *v, *e;
      if ((e=strchr(buf, '\n'))) *e = '\0';

      // get to key front
      for (k=buf; *k && isspace(*k); k++);
      if (!*k) continue;
      if (*k=='#') continue;
      // get to key end
      v = strchr(k, ' ');
      if (!v) continue;
      *v++ = '\0';
      // get to val front
      while (*v && isspace(*v)) v++;
      if (!*v) continue;
      if ((e=strchr(v, ' '))) *e = '\0';

      if (!strcasecmp(k, "AuthURL")) cfg->url = strdup(v);
      else if (!strcasecmp(k, "CertDB")) cfg->ca = strdup(v);
      else if (!strcasecmp(k, "Cert")) cfg->crt = strdup(v);
      else if (!strcasecmp(k, "Key")) cfg->key = strdup(v);
      else if (!strcasecmp(k, "Socket")) cfg->socket_path = strdup(v);
      else if (!strcasecmp(k, "ActAsHeader")) cfg->actas_hdr = strdup(v);
      else if (!strcasecmp(k, "DingService")) cfg->ding_service = strdup(v);
      else if (!strcasecmp(k, "ProxdHost")) cfg->proxd_host = strdup(v);
      else if (!strcasecmp(k, "ProxdService")) cfg->proxd_service = strdup(v);
      else if (!strcasecmp(k, "PidFile")) cfg->pid_file = strdup(v);
      else if (!strcasecmp(k, "ClientPW")) {
         FILE *pf = fopen(v,"r");
         if (pf && fgets(cfg->client_pw, 64, pf) && strlen(cfg->client_pw)>4) {
            char *s;
            if (s=strchr(cfg->client_pw, '\n')) *s = '\0';
         } else {
            syslog(LOG_ERR, "error reading client pw! %m");
            perror(v);
            return (NULL);
         }
         if (pf) fclose(pf);
      } else if (!strcasecmp(k, "Syslog")) {
         if (!strcasecmp(v, "local0")) log_facility = LOG_LOCAL0;
         else if (!strcasecmp(v, "local1")) log_facility = LOG_LOCAL1;
         else if (!strcasecmp(v, "local2")) log_facility = LOG_LOCAL2;
         else if (!strcasecmp(v, "local3")) log_facility = LOG_LOCAL3;
         else if (!strcasecmp(v, "local4")) log_facility = LOG_LOCAL4;
         else if (!strcasecmp(v, "local5")) log_facility = LOG_LOCAL5;
         else if (!strcasecmp(v, "local6")) log_facility = LOG_LOCAL6;
         else if (!strcasecmp(v, "local7")) log_facility = LOG_LOCAL7;
         else fprintf(stderr, "invalid log name (%s), will use default.", v);
      }
   }
   fclose(f);
   return (cfg);
}

/* main */
 
void usage() {
   fprintf(stderr, "usage: $s [-d <debug>] [-v <verbose>] [-n <no_daemon>] -c cfg_file\n");
   exit (1);
}


int main(argc, argv)
int argc;
char **argv;
{
    int f;
    int i;
    int len;
    socklen_t fromlen;
    struct sockaddr_un sin, fromunix;
    char pidt[12];
    int bedaemon = 1;
    char *config_file_name = NULL;
  
    struct sigaction sigactn;

    my_name = argv[0];
    while (--argc > 0) {
       argv++;
       if (argv[0][0] == '-') {
          switch (argv[0][1]) {
          case 'c':
             --argc;
             argv++;
             config_file_name = argv[0];
             break;
          case 'v':
             verbose = 1;
             break;
          case 'd':
             verbose = 2;
             break;
          case 'n':
             bedaemon = 0;
             break;
          }
        }
    }
 
    // read the config first, in case we can't get started

    if (!config_file_name) usage();
    gws_config = getconfig(config_file_name);
    if (!gws_config) exit (1);

    /* Become a daemon */

    if (bedaemon) {
      if (fork()) exit (0);
      for (f = 0; f < 5; f++) (void) close(f);
      (void) open("/dev/null", O_RDONLY);
      (void) open("/dev/null", O_WRONLY);
      (void) dup(1);
      f = open("/dev/tty", O_RDWR);
      if (f > 0) {
         ioctl(f, TIOCNOTTY, 0);
         (void) close(f);
      }
    }
 
   sigemptyset(&sigsmask);

   for (i=0;sigs_to_catch[i];i++) sigaddset(&sigsmask, sigs_to_catch[i]);
   sigprocmask(SIG_BLOCK, &sigsmask, NULL);

   sigactn.sa_handler = handle_signals;
   sigactn.sa_mask = sigsmask;
   sigactn.sa_flags = 0;
   for (i=0;sigs_to_catch[i];i++) {
      if (sigaction(sigs_to_catch[i], &sigactn, NULL) < 0 ) {
         LOG(LOG_ERR, "sigaction(%d) failed: %d", sigs_to_catch[i], errno); 
         err_exit();
      }
   }
   
    /*  Open log, check and lock the lock file */

    int log_option = LOG_PID;
    if (verbose>1) log_option |= LOG_PERROR;
    openlog(LOGNAME, log_option, log_facility);
    LOG(LOG_INFO, "gwsd version %s starting", VERSION);
    (void) umask(0);
 
    pidfd = open(gws_config->pid_file, O_WRONLY|O_CREAT, 0644);
    if (pidfd < 0) {
       LOG(LOG_ERR, "%s: %m", gws_config->pid_file);
       err_exit();
    }
    if (lockf(pidfd, F_TLOCK,0) < 0) {
       if (errno == EDEADLK) {   /* Another daemon running */
          LOG(LOG_ERR, "gwsd already running!");
          err_exit();
       }
       LOG(LOG_ERR, "%s: %m", gws_config->pid_file);
       err_exit();
    }
    ftruncate(pidfd, 0);

    sprintf(pidt, "%u\n", getpid());
    f = strlen(pidt);
    if (write(pidfd, pidt, f) != f) {
       LOG(LOG_ERR, "%s: %m", gws_config->pid_file);
       err_exit();
    }

    /* Initialize the curl dataabse */
    
    if (!gwslib_curl_init(gws_config)) {
        syslog(LOG_ERR, "could not start!");
        exit (1);
    }
    
    /* Initialize the reporting system */

    ding_init(gws_config);

    pthread_mutex_init(&client_lock, NULL);
    pthread_mutex_init(&gws_lock, NULL);
 
    /* Open our UNIX listener */

    funix = socket(AF_UNIX, SOCK_STREAM, 0);
    if (funix<0) {
       LOG(LOG_ERR, "unix socket: %m");
       err_exit();
    }
 
    memset(&sin,'\0',sizeof(sin));
    sin.sun_family = AF_UNIX;
    strncpy(sin.sun_path, gws_config->socket_path, sizeof(sin.sun_path));
    unlink(sin.sun_path);
    len = strlen(sin.sun_path) + sizeof(sin.sun_family);
    if (bind(funix, (struct sockaddr *)&sin, len) < 0) {
       LOG(LOG_ERR, "bind: %m");
       funix = (-1);
       err_exit();
    }

    listen(funix,5);

    gwsd_started = time(NULL);
    sigprocmask(SIG_UNBLOCK, &sigsmask, NULL);

    /*  Wait for new connection requests. */
    
    for (;;) {
       Client C;
       int fd;
 
       fd = 0;
       fromlen = sizeof(fromunix);
       fd = accept(funix, (struct sockaddr *)&fromunix, &fromlen);
       if (fd < 0) {
            if (errno != EINTR) LOG(LOG_ERR, "unix accept: %m");
       }
   
       /* Accept a new client */
       if (fd>0) {
           LOCK(client_lock);
           C = new_client();
           C->seq = seq++;
           C->fd = fd;
           if (verbose) LOG(LOG_DEBUG, "Connect 1 %d",fd);
           pthread_create(&C->handler, NULL, th_client_handler, (void*) C);

           if (verbose) LOG(LOG_DEBUG,"Connection %d on fd %d.\n",
              C->seq, fd);
           UNLOCK(client_lock);
       }

   } 
}
 
