/* ========================================================================
 * Copyright (c) 2011 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

/* gwsd webservice functions */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "openssl/ssl.h"
#include "openssl/x509.h"
#include "openssl/x509v3.h"
#include "openssl/err.h"

#include "gwsd.h"
#include "gwslib.h"

extern int verbose;

/* --- curl and ssl tools ---- */

char curl_error_buffer[CURL_ERROR_SIZE];
/* SSL verify callback.
   If verifying, check that the cert name matches the peer name.
   Return 1 if OK. */

char *static_peer_name = NULL;
int curl_status;

static int sslverify_callback(int ok, X509_STORE_CTX *ctx)
{
   X509 *peercert;
   char cn[256];
   STACK_OF(GENERAL_NAME) *altnams;
   int i, na;
   int d =  X509_STORE_CTX_get_error_depth(ctx);

   if (verbose>1) syslog(LOG_DEBUG, "verifying gws server certificate");
   peercert=X509_STORE_CTX_get_current_cert(ctx);
   X509_NAME_get_text_by_NID (X509_get_subject_name(peercert), NID_commonName, cn, sizeof(cn));

   if (!ok) {
      int err = X509_STORE_CTX_get_error(ctx);
      syslog(LOG_ERR, " .. not ok! cert verify: d=%d, dn=%s, err=%d %s\n", d, cn, err, X509_verify_cert_error_string(err));
      return (ok);
   }
   if (d) return (ok);
   if (!static_peer_name) return (ok);

   if (verbose>1) syslog(LOG_DEBUG, "verifying certificate names\n");

   /* Else check altnames first */

   altnams = X509_get_ext_d2i(peercert, NID_subject_alt_name, NULL, NULL);
   na = sk_GENERAL_NAME_num(altnams);
   for (i=0; i<na; i++) {
      char *altn;
      GENERAL_NAME *ck = sk_GENERAL_NAME_value(altnams, i);
      if (ck->type != GEN_DNS) continue;
      altn = (char *)ASN1_STRING_data(ck->d.ia5);
      if (!strcmp(altn, static_peer_name)) break;
   }
   GENERAL_NAMES_free(altnams);

   if (i<na) return (1); /* name ok */
   if (na<0) {  /* RFC2459: altnames must be used if present */
      if (!strcmp(cn,static_peer_name)) return (1);
   }

   /* make up an error */
   syslog(LOG_ERR, " .. peer CN does not match peer name\n");
   SSLerr(SSL_F_SSL_VERIFY_CERT_CHAIN, SSL_R_PEER_ERROR_CERTIFICATE);
   report_error(7, "gws", "GWS cert failed auth!", NULL);
   return (0);
}



static CURLcode curl_ctx_callback(CURL *curl, void *ctx, void *parm)
{
   SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, sslverify_callback);
   return (0);
}


/* curl text reciever. Ignore. We want only the status. */

static size_t page_reader(void *buf, size_t len, size_t num, void *wp)
{
  return (len*num);
}

/* curl header reader. Look for status. */

static size_t header_reader(void *buf, size_t len, size_t num, void *wp)
{
  if (verbose>1) syslog(LOG_DEBUG, "recv header (%d bytes): %s", len*num, buf);
  /* looking for Status: xxx */
  if (!strncasecmp(buf, "HTTP/1.", 7)) {
     curl_status = atoi(buf+9);
  }
  return (len*num);
}


/* New config struct, load with defaults */

int gwslib_curl_init(GWSConfig *cfg)
{
   CURL *curl;

   curl = curl_easy_init();;

   curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 1);
   curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 1);
   curl_easy_setopt(curl, CURLOPT_SSL_CTX_FUNCTION, curl_ctx_callback);

   curl_easy_setopt(curl, CURLOPT_SSLCERT, cfg->crt);
   curl_easy_setopt(curl, CURLOPT_SSLCERTTYPE, "PEM");
   curl_easy_setopt(curl, CURLOPT_SSLKEY, cfg->key);
   curl_easy_setopt(curl, CURLOPT_SSLKEYTYPE, "PEM");
   curl_easy_setopt(curl, CURLOPT_CAINFO, cfg->ca);

   curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, page_reader);
   curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, header_reader);
   curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, curl_error_buffer);

   if (verbose>1) curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);

   curl_easy_setopt(curl, CURLOPT_HTTPGET, 1);

   cfg->curl = curl;

   // extract the host name from the url

   if (!strncmp(cfg->url, "http://", 7)) static_peer_name = strdup(cfg->url+7);
   else if (!strncmp(cfg->url, "https://", 8)) static_peer_name = strdup(cfg->url+8);
   if (static_peer_name) {
      char *s;
      if (s=strchr(static_peer_name, ':')) *s = '\0';
      if (s=strchr(static_peer_name, '/')) *s = '\0';
      if (verbose>1) syslog(LOG_DEBUG, "peername: %s", static_peer_name);
   }

   return (1);
}


/* ---- Exported group membership functions ----------------- */


/* Test if an id is an effective member of a group. 
   Returns 1 if yes, 0 if not.
 */

int gwslib_is_member(GWSConfig *cfg, char *group, char *id, char *actas) 
{
   char url[1024];
   struct curl_slist *headers = NULL;

   snprintf(url, 1024, cfg->url, group, id);
   if (!cfg->curl) gwslib_curl_init(cfg);

   if (actas && *actas) {
      char *hdr = (char*) malloc(strlen(actas)+16);
      char *s;
      if (s=strchr(actas, '\n')) *s = '\0';
      sprintf(hdr, "%s: %s", cfg->actas_hdr, actas);
      struct curl_slist *headers = curl_slist_append(NULL, hdr);
      curl_easy_setopt(cfg->curl, CURLOPT_HTTPHEADER, headers);
      free(hdr);
   }

   if (verbose>1) syslog(LOG_DEBUG, "gws req: %s", cfg->url);

   curl_easy_setopt(cfg->curl, CURLOPT_URL, url);
   curl_status = 0;
   curl_easy_perform(cfg->curl);
   if (headers) curl_slist_free_all(headers);
   if (curl_status==0) report_error(7, "gws", "GWS request failed!", url);
   return curl_status;
}
       

