/* ========================================================================
 * Copyright (c) 2011 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */


/* UW GWS authorization module (groups membership) */

#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>


#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_core.h"

typedef apr_pool_t pool;
typedef apr_table_t table;

#include "gwsd.h"
#include "gwslib.h"
#include "gwscrypt.h"

/* default cookie name */
#define COOKIE   "modgws"
#define GWS_EPPN_DEFAULT "@washington.edu"

module gws_module;

#include "apr_strings.h"

#define MY_LOG_DEBUG APLOG_MARK,APLOG_DEBUG,0
#define MY_LOG_INFO  APLOG_MARK,APLOG_INFO,0
#define MY_LOG_ERR   APLOG_MARK,APLOG_ERR,0
#define MY_LOG_EMERG APLOG_MARK,APLOG_EMERG,0

// define for really debugging
#define MOD_GWS_DEBUG
#if defined(MOD_GWS_DEBUG)
#define DEBUG0(t) ap_log_error(MY_LOG_DEBUG, gws_main_server, t)
#define DEBUG1(t,a1) ap_log_error(MY_LOG_DEBUG, gws_main_server, t, a1)
#define DEBUG2(t,a1, a2) ap_log_error(MY_LOG_DEBUG, gws_main_server, t, a1, a2)
#else
#define DEBUG0(t) 
#define DEBUG1(t,a1) 
#define DEBUG2(t,a1, a2) 
#endif

/* gws configuration */

// globals
char *gwsd_sock_name;
char *gwsd_sock_pw;
char *gws_cookie_name = "gwssession";
int owner_access = 0;
int timeout_min = 9*60;
char *gws_keyword = "group";
char *ug_keyword = "ugroup";
char *user_keyword = "user";

// access to gwsd is one at a time
apr_thread_mutex_t *gwsd_lock;
int gwsd_sock;
int gwsd_seq = 1;

server_rec *gws_main_server;  // used for some logging

/* Apache 2.2.x headers must be accumulated and set in the output filter.
   Apache 2.0.49+ also supports the filters as well.
   In Apache 1.3.x and lesser 2.0.x we write the headers directly. */

#if AP_MODULE_MAGIC_AT_LEAST(20020903,6)
#define GWS_DEFERRED_HEADERS

/* Append add the entries in 'src' to the 'dest' table */
static void append_to_table(request_rec *r, apr_table_t *dest, apr_table_t *src)
{
   const apr_array_header_t *srce = apr_table_elts(src);
   int i;

   for (i=0; i<srce->nelts; i++) {
      apr_table_entry_t *ent = &((apr_table_entry_t *) (srce->elts))[i];
      ap_log_rerror (MY_LOG_DEBUG, r, "gws adding header %s", ent->key);
      apr_table_add(dest, ent->key, ent->val);
   }
}
#endif

/* request record hashes */

#define MAX_CACHE 40     // max cached request line hashes
#define MAX_CTEXT 180    // MAX_CACHE*sizeof(u_int)+20
#define MAX_ID_LEN 128   // max id length

/* server config */
typedef struct {
   int owner_access;  /* if 1, access gws as resource owner  */
   int authoritative; 
} gws_config_rec;

/* per request record */
typedef struct {
  int userok;
  u_int uhash[MAX_CACHE];   // auth cache
  int nuhash;
#ifdef GWS_DEFERRED_HEADERS
   table *headers_out;
#endif
} gws_request_rec;

/* per dir conf */
typedef struct {
   int active;
   int authoritative;
} gws_dir_config_rec;


/* ------- remote gwsd functions --------------- */


/* send msg and get resp, caller has gwsd lock */
int connect_to_gwsd();
int gwsd_send_receive(ClientMsg msg, ResponseMsg rsp)
{
   int i;
  for (i=0; i<3; i++) {
   msg->seq = gwsd_seq++;
   DEBUG1("send to gwsd %d", msg->cmd);
   int nb = write(gwsd_sock, (char*)msg, sizeof(ClientMsg_));
   // DEBUG1("gws_gwsd_write: %d", errno);
   if (nb<0 && errno==EPIPE) {
      DEBUG0("reconnecting");
      gwsd_sock = 0;
      connect_to_gwsd();
      continue;  // try again
   }
   if (nb!=sizeof(ClientMsg_)) {
     ap_log_error(MY_LOG_ERR, gws_main_server, "gws_gwsd_write: %d", errno);
     return (0);
   }

   nb = read(gwsd_sock, rsp, sizeof(ResponseMsg_));
   if (nb!=sizeof(ResponseMsg_)) {
      ap_log_error(MY_LOG_ERR, gws_main_server, "gws_gwsd_read: %d", errno);
      gwsd_sock = 0;
      connect_to_gwsd();
      continue;
   }
   if (rsp->seq!=msg->seq) {
      ap_log_error(MY_LOG_ERR, gws_main_server, "gws_gwsd_seq: %d should be %d", rsp->seq, msg->seq);
      gwsd_sock = 0;
      connect_to_gwsd();
      continue;
   }
   // DEBUG0(".. ok");
   return (1);
  }
  return (0);
}

/* connect if not already connected, have gwsd lock */
int connect_to_gwsd()
{
   struct sockaddr_un sun;
   int ret;

   if (gwsd_sock) return (1);
   if (!gwsd_sock_pw) {
      ap_log_error(MY_LOG_ERR, gws_main_server, "No socket password configured.  [GWSpasswordfile needed] ");
      return (0);
   }
   sun.sun_family = AF_UNIX;
   strcpy(sun.sun_path, gwsd_sock_name);
   if ((gwsd_sock=socket(AF_UNIX, SOCK_STREAM, 0))>0 &&
      connect(gwsd_sock, (struct sockaddr *)&sun, strlen(sun.sun_path)+2)==0) {
      ap_log_error(MY_LOG_DEBUG, gws_main_server, "gws_gwsd_connect ok");
   } else if (gwsd_sock) {
      close(gwsd_sock);
      gwsd_sock = 0;
      ap_log_error(MY_LOG_ERR, gws_main_server, "gws_gwsd_connect: %d", errno);
      return (0);
   }
   ClientMsg_ msg;
   ResponseMsg_ rsp;
   rsp.resp = 900;

   msg.cmd = CMD_PASSWORD;
   strncpy((char*)msg.id, gwsd_sock_pw, MSG_ID_LEN);
   ret = gwsd_send_receive(&msg, &rsp);
   DEBUG1("connected %d", ret);
   return (ret);
}

/* test membership */
int remote_is_member(char *group, char *user, char *actas)
{
   ClientMsg_ msg;
   ResponseMsg_ rsp;

   apr_thread_mutex_lock(gwsd_lock);

   if (!connect_to_gwsd()) {
     apr_thread_mutex_unlock(gwsd_lock);
     return (0);
   }

   DEBUG2("chk %s in %s", user, group);
   msg.cmd = CMD_MEMBER_OF;
   strncpy((char*)msg.id, user, MSG_ID_LEN);
   strncpy((char*)msg.group, group, MSG_GROUP_LEN);
   if (actas) {
      strncpy((char*)msg.actas, actas, MSG_ACTAS_LEN);
      DEBUG1("actas %s", actas);
   } else msg.actas[0] = '\0';
   int ret = gwsd_send_receive(&msg, &rsp);

   apr_thread_mutex_unlock(gwsd_lock);

   DEBUG2("rsp %d (%d)", rsp.resp, ret);
   if (ret && rsp.resp==200) return (1);
   return (0);
}


/* ----  module configuration ---- */


// set cookie name
const char *set_gws_cookie_name (cmd_parms *cmd, void *dummy, const char *opt1)
{
    if (strlen(opt1)<63) {
       gws_cookie_name = apr_pstrdup(cmd->pool, opt1);
       DEBUG1("gws cookie name = %s", opt1);
    }
    return NULL;
}

// set socket name
const char *set_gws_socket_name (cmd_parms *cmd, void *dummy, const char *opt1)
{
   gwsd_sock_name = apr_pstrdup(cmd->pool, opt1);
   ap_log_error(MY_LOG_DEBUG, gws_main_server, "gws_gwsd_sock_name: %s", opt1);
   return NULL;
}

// set socket key
const char *set_gws_pw (cmd_parms *cmd, void *dummy, const char *opt1)
{
   char pw[64];
   FILE *f = fopen(opt1,"r");
   if (f && fgets(pw, 64, f) && strlen(pw)>4) {
      char *s;
      gwsd_sock_pw = apr_pstrdup(cmd->pool, pw);
      if (s=strchr(gwsd_sock_pw, '\n')) *s = '\0';
      ap_log_error(MY_LOG_DEBUG, gws_main_server, "gws_gwsd_pw_ok: %s", opt1);
   } else ap_log_error(MY_LOG_ERR, gws_main_server, "read gwsd pw failed! %d", errno);
   if (f) fclose(f);
   return NULL;
}

// set owner access
const char *set_gws_owner_access (cmd_parms *cmd, void *dummy, int flag)
{
  owner_access = flag;
  ap_log_error(MY_LOG_DEBUG, gws_main_server, "owner_access: %d", flag);
  return NULL;
}

// set timeout 
const char *set_gws_timeout (cmd_parms *cmd, void *dummy, const char *opt1)
{
   timeout_min = atoi(opt1);
   ap_log_error(MY_LOG_DEBUG, gws_main_server, "gws_timeout: %d minutes", timeout_min);
   return NULL;
}

// set gws keyword
const char *set_gws_keyword (cmd_parms *cmd, void *dummy, const char *opt1)
{
   gws_keyword = apr_pstrdup(cmd->pool, opt1);
   ap_log_error(MY_LOG_DEBUG, gws_main_server, "gws_keyword: %s", opt1);
   return NULL;
}

// set unix group keyword
const char *set_ug_keyword (cmd_parms *cmd, void *dummy, const char *opt1)
{
   ug_keyword = apr_pstrdup(cmd->pool, opt1);
   ap_log_error(MY_LOG_DEBUG, gws_main_server, "ug_keyword: %s", opt1);
   return NULL;
}

// set user keyword
const char *set_user_keyword (cmd_parms *cmd, void *dummy, const char *opt1)
{
   user_keyword = apr_pstrdup(cmd->pool, opt1);
   ap_log_error(MY_LOG_DEBUG, gws_main_server, "user_keyword: %s", opt1);
   return NULL;
}


// set authoritative
const char *set_gws_authoritative (cmd_parms *cmd, void *dir_confv, int flag)
{
  void *server_conf = cmd->server->module_config;
  gws_config_rec *gc = (gws_config_rec *) ap_get_module_config(server_conf, &gws_module);
  gc->authoritative = flag;
  return NULL;
}



/* ---- process entry points ------------ */

static int gws_init(pool *p, pool *plog, pool *ptemp, server_rec *s) {
   ap_log_error(MY_LOG_INFO, s, "mod_gws init");
   ap_add_version_component(p, PACKAGE_NAME "/" PACKAGE_VERSION);
   if (gwsd_sock_pw) gws_crypt_init(gwsd_sock_pw, strlen(gwsd_sock_pw));
   else ap_log_error(MY_LOG_ERR, gws_main_server, "No socket password configured.  [GWSpasswordfile needed] ");
   gws_main_server = s;
   apr_thread_mutex_create(&gwsd_lock, APR_THREAD_MUTEX_DEFAULT, p);
   return OK;
}

static int gws_pre_config(pool *p, pool *plog, pool *ptemp) {
  DEBUG0("gws_pre_config");
  // gws_crypt_init();
  return OK;
}

/* Start of a request.  If using filters make the request record. */

static int gws_post_read (request_rec * r)
{
    gws_request_rec *rr = apr_pcalloc(r->pool, sizeof(gws_request_rec));
    
    ap_log_rerror(MY_LOG_DEBUG, r, "gws_post_read");
    rr->userok = 0;
#ifdef GWS_DEFERRED_HEADERS
    ap_set_module_config (r->request_config, &gws_module, rr);
    rr->headers_out = apr_table_make(r->pool, 3);

    /* If this is a forwarded request, copy our headers from the parent. */
    if (r->prev) {
       int n;
       gws_request_rec *prr = (gws_request_rec *) ap_get_module_config (r->prev->request_config,
                                                     &gws_module);
       if (prr) {
          n = apr_table_elts(prr->headers_out)->nelts;
          ap_log_rerror (MY_LOG_DEBUG, r, "gws of: forwarding %d output headers to redirect request", n);
          if (n) append_to_table(r, rr->headers_out, prr->headers_out);

          rr->userok = prr->userok;
       }
    }

#endif

    return DECLINED;
}


static void *create_gws_config (pool *p, struct server_rec *d)
{
  gws_config_rec *gc = (gws_config_rec *) apr_pcalloc (p, sizeof (gws_config_rec));

  DEBUG0("<create_gws_config>");
  gws_crypt_seed();
  gc->owner_access = 0;
  ap_log_error(MY_LOG_INFO, d, "create_gws_config");
  return (gc);
}

static void *create_gws_dir_config (pool * p, char *d)
{
  gws_dir_config_rec *gdc = (gws_dir_config_rec *) apr_pcalloc (p, sizeof (gws_dir_config_rec));

  DEBUG0("<create_gws_dir_config>");
  gws_crypt_seed();
  gdc->authoritative = 0;
  return (gdc);
}



/* Output filter tools */

#ifdef GWS_DEFERRED_HEADERS

static void set_output_filter(request_rec *r)
{
   DEBUG0("gws adding output filter");
   ap_add_output_filter("GWS_HEADERS_OUT", NULL, r, r->connection);
}

static apr_status_t do_output_filter(ap_filter_t *f,
                                             apr_bucket_brigade *in)
{
    request_rec *r = f->r;
    gws_request_rec *rr = (gws_request_rec *) ap_get_module_config (r->request_config,
                                                     &gws_module);

    if ( rr ) {
        ap_log_rerror (MY_LOG_DEBUG, r, "gws output_filter: merging %d output headers",
                   apr_table_elts(rr->headers_out)->nelts);
        append_to_table(r, r->headers_out, rr->headers_out);
    }

    /* remove ourselves from the filter chain */
    ap_remove_output_filter(f);

    /* send the data up the stack */
    return ap_pass_brigade(f->next,in);
}
#endif /* GWS_DEFERRED_HEADERS */



/* Check authorization */

// strip local eppn scope
static char *strip_user(char *u)
{
   char *ret;
   char *p;
   
   if (!u) return (NULL);

   ret = strdup(u);
   if ((p=strstr(ret,GWS_EPPN_DEFAULT))) *p = '\0';
   return (ret);
}
      
/* cookie: rand(8) + 'GWS ' + time(4) + id + '\0' + cache hashes  */

int gws_set_cookie (request_rec *r, u_int *hash, int nhash, u_int t_min)
{

#ifdef GWS_DEFERRED_HEADERS
   gws_request_rec *rr = (gws_request_rec*)  ap_get_module_config (r->request_config, &gws_module);
   if (!rr) return (0);
#endif

   char ctext[MAX_CTEXT], *etext, *b64;
   int ctextl, etextl;
   time_t t = time(NULL);
   char ck[1024];
   int i;
   char *ruser = strip_user(r->user);
   int ruserl = strlen(ruser);
   ctextl = 16 + ruserl + 1;
   memset(ctext, 0, ctextl);
   if (t_min==0) t_min = (t/60) + timeout_min;

   RAND_bytes(ctext, 8);
   strcpy(ctext+8, "GWS ");
   memcpy(ctext+12, &t_min, 4);
   strcpy(ctext+16, ruser);
   ctext[16+ruserl] = '\0';
   // add the hashes
   memcpy(ctext+ctextl, &nhash, sizeof(int));
   ctextl += sizeof(int);
   for (i=0;i<nhash;i++) {
      memcpy(ctext+ctextl,&hash[i],sizeof(u_int));
      ctextl += sizeof(u_int);
   }

   DEBUG2("gws: set cookie, user=%s, %d hashes", ruser, nhash);

   gws_crypt(GWS_ENCRYPT, &etext, &etextl, ctext, ctextl);
   gws_data_to_b64(&b64, &ctextl, etext, etextl);

   sprintf(ck, "%s=%s; path=/; secure", gws_cookie_name, b64);
   free(etext);
   free(b64);
   DEBUG1("  set-cookie: %s", ck);
#ifdef GWS_DEFERRED_HEADERS
   apr_table_add(rr->headers_out, "Set-Cookie", ck);
#else
   apr_table_add(r->headers_out, "Set-Cookie", ck);
#endif
   return (1);
}

u_int gws_get_cookie (request_rec *r, u_int *hash, int *nhash)
{

   char *ck, *ckp;
   char *value;
   char *etext;
   int ctextl, etextl;
   u_int ct_min = 0;
   time_t nt = time(NULL);
   u_int nt_min = nt/60;
   char *ctext;
   char *ruser = strip_user(r->user);
   int n = 0;
   int i,p;

   DEBUG0("gws: cookie check");

   /* Look for the login cookie */

   if (!(ckp=(char*)apr_table_get (r->headers_in, "Cookie"))) return 0;

   /* search the cookies */
   if (!(ck = apr_palloc (r->pool, 2 + strlen (ckp)))) {
       ap_log_rerror(MY_LOG_ERR, r, "mod_gws: palloc failed!");
       return 0;
   };
   strcpy (ck, ckp);
   ck[0 + strlen (ckp)] = ';';
   ck[1 + strlen (ckp)] = '\0';

   /* be wary of correctly named, bogus cookies */
   for (ck=strtok (ck," ;\n\r\t\f"); ck; ck=strtok (NULL, " ;\n\r\t\f")) {

      if (!ck) break;
      if (!(value = strchr (ck,'='))) continue;

      *value = '\0';
      value++;

      if (strcmp(ck, gws_cookie_name)) continue;

      /* is ours, value is encrypted stuff */

      DEBUG1("gws: ck val = [%s]", value);
         

      gws_b64_to_data(&etext, &etextl, value, strlen(value)); 
      gws_crypt(GWS_DECRYPT, &ctext, &ctextl, etext, etextl);
      free(etext);

      if (strncmp(ctext+8,"GWS ",4)) {
         DEBUG1("gws: bogus cookie: not GWS, c=%d", ctextl);
         free(ctext);
         continue;
      }

      memcpy(&ct_min, ctext+12, 4);
      DEBUG2("cookie hour = %d, now = %d", ct_min, nt_min);
      DEBUG1("cookie user = %s", ctext+16);
      if (nt_min>ct_min) {
         DEBUG2("gws: cookie expired=%d>%d", nt_min, ct_min);
      } else if (strcmp(ctext+16, ruser)) {
         DEBUG2("gws: cookie wrong user=%s:%s", ctext+16, ruser);
      } else {
         p = 16 + strlen(ruser) + 1;
         memcpy(&n, ctext+p, sizeof(int));
         DEBUG1("gws: cookie ok, nhash=%d", n);
         p += sizeof(int);
         for (i=0; i<n; i++) {
            memcpy(&(hash[i]), ctext + p, sizeof(u_int));
            DEBUG1("gws: got hash=%x", hash[i]);
            p += sizeof(u_int);
         }
         *nhash = n;
         free(ctext);
         return ct_min;
      }
      free(ctext);
      return (0);
   }
   return (0);
}

/* test id user is in a unix group */

int ugroup_is_member(char *group_name, char *user)
{
    struct group *grp;
    int ret = 0;
    int gid;
    char **gent;

    if (!(grp=getgrnam(group_name))) {
       ap_log_error(MY_LOG_ERR, gws_main_server, "group %s not found", group_name);
       return 0;
    }
    setgrent();
    gid = grp->gr_gid;
    DEBUG2("gid=%d user=%s", gid, user);
    while (grp=getgrent()) {
        if (gid != grp->gr_gid) continue;
        DEBUG1("checking %s", grp->gr_name);
        for (gent=grp->gr_mem; *gent && **gent; ++gent) {
           if (!strcmp (user, *gent)) {
              DEBUG0("got it");
              ret = 1;
              break;
           }
        }
        if (ret) break;
    }
    endgrent();
    return ret;
}
   

/* Evaluate a require line */


#define NO_MATCHER 0
#define GWS_MATCHER 1
#define UG_MATCHER 2
#define USER_MATCHER 3

#define OP_OR 0
#define OP_AND 1

// get next token
char *next_token(apr_pool_t *ap, const char **line) 
{
   const char *pos = *line;
   int len;
   char *res;
  
   if (!*pos) return NULL;

   if (*pos=='(' || *pos==')') {
      ++pos;
      len = 1;
   } else {
      while (*pos && !(*pos=='(' || *pos==')' || isspace(*pos)) ) ++pos;
      len = pos - *line;
   }
   res = (char *)apr_palloc(ap, len+1);
   memcpy(res, *line, len);
   res[len] = 0;

   while (isspace(*pos) ) ++pos;
   *line = pos;

   return res;
}

/* The require line:  keys = 'user' 'group' 'ugroup'
 *   1. '()' delimit subrequires
 *   2. Assume 'or' for main and subrequires ( e.g. user u1 u2 u3 => user u1 or user u2 or user u3 )
 *
 * Returns:  -1 = fails, 1 = passes, 0 = not ours
 */

int evaluate_requirement(request_rec *r, const char **str, char *ruser, char *actas)
{
   char *w;
   int sofar = 0;
   int op = OP_OR; // OR
   int matcher = 0;
   int neg = 0;

   DEBUG1("eval: %s", *str);

   while (w = next_token(r->pool, str)) {
      
      if (!strcasecmp(w, "and")) {
         if (sofar==0) return (0);
         if (sofar<0) {
            DEBUG0("  AND with false fails");
            break;
         }
         op = OP_AND;
         continue;
         
      } else if (!strcasecmp(w, "or")) {
         if (sofar==0) return (0);
         if (sofar>0) {
            DEBUG0("  OR with true passes");
            break;
         }
         op = OP_OR;
         continue;

      } else if (!strcasecmp(w, "not")) {
         neg = 1;
         continue;

      } else if (!strcasecmp(w, ")")) {
         DEBUG1("  leave sub, sofar=%d", sofar);
         return (sofar);
      }
     
      // see if we can stop now
      if (sofar==1 && op==OP_OR) break;
      if (sofar==(-1) && op==OP_AND) break;

      if (!strcasecmp(w, gws_keyword)) {
         matcher = GWS_MATCHER;
         
      } else if (!strcasecmp(w, ug_keyword)) {
         matcher = UG_MATCHER;
         
      } else if (!strcasecmp(w, user_keyword)) {
         matcher = USER_MATCHER;

      } else {

         int v = 0;
         if (!strcasecmp(w, "(")) {
            v = evaluate_requirement(r, str, ruser, actas);
            if (v<0) v = 0;

         } else if (matcher==GWS_MATCHER) {
            v = remote_is_member((char*)w, ruser, actas);

         } else if (matcher==UG_MATCHER) {
            v = ugroup_is_member((char*)w, ruser);

         } else if (matcher==USER_MATCHER) {
            char *p = strstr(w,GWS_EPPN_DEFAULT);  // already stripped from userid
            if (p) *p = '\0';
            v = !strcmp((char*)w, ruser);

         } else return (0);  // not for us, decline

         if (neg) v = (1-v);
         // DEBUG1("  eval v=%d", v);
         if (v) sofar = 1;
         else sofar = (-1);
      
      }
       
   }
   // skip to end of present section
   int depth = 0;
   while (w = next_token(r->pool, str)) {
      DEBUG1("  skip [%s]", w);
      if (*w=='(') depth++;
      if (*w==')') depth--;
      if (depth<0) break;
   }
   
   DEBUG1("eval end, val=%d", sofar);

   return (sofar);
}


/* Check authorization */

static int gws_authorize (request_rec * r)
{
   gws_dir_config_rec *gdc = (gws_dir_config_rec *) ap_get_module_config (r->per_dir_config, &gws_module);
   gws_request_rec *rr = (gws_request_rec *) ap_get_module_config (r->request_config, &gws_module);

   char *ruser = strip_user(r->user);
   int m = r->method_number;
   const apr_array_header_t *reqs_arr = ap_requires(r);
   require_line *reqs;
   const char *t;
   int i;
   char *actas = NULL;
   char badpwent = 0;
   u_int *uhash = NULL;
   int *nuhash = NULL;
   gws_request_rec *prr = NULL;
   u_int rhash;
   int newhash = 0;
   int result = 0;
   u_int c_min = 0;

   DEBUG1("gws_authorize: %d", gdc->active);

   if (!reqs_arr) return DECLINED;
   if (!gdc->active) return DECLINED;

   DEBUG0("gws_authorize 3");

   // get the session cookie
   if (r->main) {
      prr = (gws_request_rec *) ap_get_module_config (r->main->request_config, &gws_module);
      if (prr) {
         uhash = prr->uhash;
         nuhash = &prr->nuhash;
      } else if (rr) {
         uhash = rr->uhash;
         nuhash = &rr->nuhash;
      }
   } else if (rr) {
      c_min = gws_get_cookie(r, rr->uhash, &rr->nuhash);
      uhash = rr->uhash;
      nuhash = &rr->nuhash;
   }
   
   // see if we need 'act-as' checking
   if (owner_access && r->finfo.user>0) {
      struct passwd *pw = getpwuid(r->finfo.user);
      if (pw) actas = apr_pstrdup(r->pool, pw->pw_name);
      else {
         ap_log_rerror(MY_LOG_ERR, r, "No pwent for %d !", r->finfo.user);
         badpwent = 1;
      }
   }

   // scan the requires
   reqs = (require_line *)reqs_arr->elts;
 
   for (i=0; i<reqs_arr->nelts; i++) {
      int h;
      if (!(reqs[i].method_mask & (AP_METHOD_BIT << m))) {
          continue;
      }
      t = reqs[i].requirement;
      DEBUG1("  check require [%s]", t);
      
      rhash = gws_crypt_hash((char*)t);
      for (h=0; nuhash && h<*nuhash; h++) {
         if (rhash==uhash[h]) {
            DEBUG1("  found hash %x", rhash);
            result = 1;
            break;
         }
      }
      if (result==1) break;

      result = evaluate_requirement(r, &t, ruser, actas); 
      // update cookie
      if (result>0 && nuhash) {
         if (*nuhash>MAX_CACHE-2) {
            *nuhash=0; /* cache full - just clear and start over */
         }
         uhash[*nuhash] = rhash;
         *nuhash += 1;
         newhash++;
      }
      if (result==1) break;
      
   }
 
   // no access if we're authoritative
   if (result==0) return DECLINED;
   if (result<0 && gdc->authoritative) {
      DEBUG0("no passing requires and authoritative: FAIL");
      ap_note_auth_failure(r);
      return HTTP_UNAUTHORIZED;
   }

   if (newhash && !r->main) gws_set_cookie(r, uhash, *nuhash, c_min);
 
   if (result>0) return OK;

   DEBUG0("no passing requires: DECLINE");

   return DECLINED;

}


/* Authtypes for activation.  */

const char *activate_gws_dir_config (cmd_parms *cmd,
      void *dir_confv, const char *arg)
{
  gws_dir_config_rec *dir_conf = (gws_dir_config_rec*) dir_confv;
  if (!strcasecmp(arg,"uwnetid")) dir_conf->active = 1;
  else if (!strcasecmp(arg,"securid"))  dir_conf->active = 1;
  else if (!strcasecmp(arg,"shibboleth"))  dir_conf->active = 1;
  ap_log_error(MY_LOG_DEBUG, gws_main_server, "mod_gws %sactivated by %s", dir_conf->active?"":"not ", arg);
  return DECLINE_CMD;
} 

const char *deactivate_gws_dir_config (cmd_parms *cmd,
      void *dir_confv, const char *arg)
{
  gws_dir_config_rec *dir_conf = (gws_dir_config_rec*) dir_confv;
  dir_conf->active = 0;
  ap_log_error(MY_LOG_DEBUG, gws_main_server, "mod_gws deactivated by %s", arg);
  return DECLINE_CMD;
}


/* ------  Configuration ---------- */

static const command_rec gws_cmds[] =
{
     
  AP_INIT_TAKE1("AuthType",
      activate_gws_dir_config, NULL, OR_AUTHCFG,
     "Authorization type"
  ),

  AP_INIT_FLAG("GWSActive", ap_set_flag_slot,
      (void *)APR_OFFSETOF(gws_dir_config_rec, active), OR_AUTHCFG,
     "Set off to deactivate mod_gws"
  ),
     
  AP_INIT_RAW_ARGS("AuthGroupFile",
      deactivate_gws_dir_config, NULL, OR_AUTHCFG,
     "group file: deactivates mod_gws"
  ),

  AP_INIT_TAKE1("GWSCookie",
      set_gws_cookie_name, NULL, RSRC_CONF,
     "cookie_name"
  ),
     
  AP_INIT_FLAG("GWSAuthoratative", ap_set_flag_slot,
      (void *)APR_OFFSETOF(gws_dir_config_rec, authoritative), OR_AUTHCFG,
     "Use GWSAuthoritative instead"
  ),
     
  AP_INIT_FLAG("GWSAuthoritative", ap_set_flag_slot,
      (void *)APR_OFFSETOF(gws_dir_config_rec, authoritative), OR_AUTHCFG,
     "Set on to require mod_gws success"
  ),
     
  AP_INIT_TAKE1("GWSsocketname",
      set_gws_socket_name, NULL, RSRC_CONF,
     "gwsd socket name"
  ),
     
  AP_INIT_TAKE1("GWSpasswordfile",
      set_gws_pw, NULL, RSRC_CONF,
     "gwsd password file"
  ),
     
  AP_INIT_FLAG("GWSOwnerAccess",
     set_gws_owner_access, NULL, RSRC_CONF,
    "Set on to resource owner as act-as to GWS"
  ),

  AP_INIT_TAKE1("GWStimeout",
      set_gws_timeout, NULL, RSRC_CONF,
     "session lifetime in minutes (defautl 9 hours)"
  ),
     
  AP_INIT_TAKE1("GWSGroupKeyword",
      set_gws_keyword, NULL, RSRC_CONF,
     "keyword to indicate our require line gws group check (default 'group')"
  ),
     
  AP_INIT_TAKE1("GWSUnixGroupkeyword",
      set_ug_keyword, NULL, RSRC_CONF,
     "keyword to indicate our require line unix group check (default 'ugroup')"
  ),
     
  AP_INIT_TAKE1("GWSUserkeyword",
      set_user_keyword, NULL, RSRC_CONF,
     "keyword to indicate our require line user id check (default 'user')"
  ),
     
  {NULL}
};

static void register_hooks(pool *p) {
#ifdef GWS_DEFERRED_HEADERS
    ap_register_output_filter("GWS_HEADERS_OUT", do_output_filter,
                              NULL, AP_FTYPE_CONTENT_SET);
    ap_hook_insert_filter(set_output_filter, NULL, NULL, APR_HOOK_LAST);
#endif

    ap_hook_post_config(gws_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_auth_checker(gws_authorize, NULL, NULL, APR_HOOK_REALLY_FIRST);
    ap_hook_post_read_request (gws_post_read, NULL, NULL, APR_HOOK_MIDDLE);
    // ap_hook_fixups (gws_fixups, NULL, NULL, APR_HOOK_MIDDLE);

    ap_hook_pre_config (gws_pre_config, NULL, NULL, APR_HOOK_MIDDLE);

}

module AP_MODULE_DECLARE_DATA gws_module = {
    STANDARD20_MODULE_STUFF,
    create_gws_dir_config,
    NULL,
    create_gws_config,
    NULL,   /* merge_gws_config, */
    gws_cmds,
    register_hooks,
};
